{ This file is generated by dev/parseunicode/parseunicodeclasses program }
type
  TArabicJoin = (arNone, arInitial, arMedial, arFinal, arIsolated);
  TUTF8Decomposition = record
    de, re: string; //decomposed, recomposed UTF8
    join: TArabicJoin;
  end;
const
  UTF8Decomposition : array[0..1219] of TUTF8Decomposition = (
  (de:'≮'; re:'≮'; join:arNone),
  (de:'≠'; re:'≠'; join:arNone),
  (de:'≯'; re:'≯'; join:arNone),
  (de:'À'; re:'À'; join:arNone),
  (de:'Á'; re:'Á'; join:arNone),
  (de:'Â'; re:'Â'; join:arNone),
  (de:'Ã'; re:'Ã'; join:arNone),
  (de:'Ā'; re:'Ā'; join:arNone),
  (de:'Ă'; re:'Ă'; join:arNone),
  (de:'Ȧ'; re:'Ȧ'; join:arNone),
  (de:'Ä'; re:'Ä'; join:arNone),
  (de:'Ả'; re:'Ả'; join:arNone),
  (de:'Å'; re:'Å'; join:arNone),
  (de:'Ǎ'; re:'Ǎ'; join:arNone),
  (de:'Ȁ'; re:'Ȁ'; join:arNone),
  (de:'Ȃ'; re:'Ȃ'; join:arNone),
  (de:'Ạ'; re:'Ạ'; join:arNone),
  (de:'Ḁ'; re:'Ḁ'; join:arNone),
  (de:'Ą'; re:'Ą'; join:arNone),
  (de:'Ḃ'; re:'Ḃ'; join:arNone),
  (de:'Ḅ'; re:'Ḅ'; join:arNone),
  (de:'Ḇ'; re:'Ḇ'; join:arNone),
  (de:'Ć'; re:'Ć'; join:arNone),
  (de:'Ĉ'; re:'Ĉ'; join:arNone),
  (de:'Ċ'; re:'Ċ'; join:arNone),
  (de:'Č'; re:'Č'; join:arNone),
  (de:'Ç'; re:'Ç'; join:arNone),
  (de:'Ḋ'; re:'Ḋ'; join:arNone),
  (de:'Ď'; re:'Ď'; join:arNone),
  (de:'Ḍ'; re:'Ḍ'; join:arNone),
  (de:'Ḑ'; re:'Ḑ'; join:arNone),
  (de:'Ḓ'; re:'Ḓ'; join:arNone),
  (de:'Ḏ'; re:'Ḏ'; join:arNone),
  (de:'È'; re:'È'; join:arNone),
  (de:'É'; re:'É'; join:arNone),
  (de:'Ê'; re:'Ê'; join:arNone),
  (de:'Ẽ'; re:'Ẽ'; join:arNone),
  (de:'Ē'; re:'Ē'; join:arNone),
  (de:'Ĕ'; re:'Ĕ'; join:arNone),
  (de:'Ė'; re:'Ė'; join:arNone),
  (de:'Ë'; re:'Ë'; join:arNone),
  (de:'Ẻ'; re:'Ẻ'; join:arNone),
  (de:'Ě'; re:'Ě'; join:arNone),
  (de:'Ȅ'; re:'Ȅ'; join:arNone),
  (de:'Ȇ'; re:'Ȇ'; join:arNone),
  (de:'Ẹ'; re:'Ẹ'; join:arNone),
  (de:'Ȩ'; re:'Ȩ'; join:arNone),
  (de:'Ę'; re:'Ę'; join:arNone),
  (de:'Ḙ'; re:'Ḙ'; join:arNone),
  (de:'Ḛ'; re:'Ḛ'; join:arNone),
  (de:'Ḟ'; re:'Ḟ'; join:arNone),
  (de:'Ǵ'; re:'Ǵ'; join:arNone),
  (de:'Ĝ'; re:'Ĝ'; join:arNone),
  (de:'Ḡ'; re:'Ḡ'; join:arNone),
  (de:'Ğ'; re:'Ğ'; join:arNone),
  (de:'Ġ'; re:'Ġ'; join:arNone),
  (de:'Ǧ'; re:'Ǧ'; join:arNone),
  (de:'Ģ'; re:'Ģ'; join:arNone),
  (de:'Ĥ'; re:'Ĥ'; join:arNone),
  (de:'Ḣ'; re:'Ḣ'; join:arNone),
  (de:'Ḧ'; re:'Ḧ'; join:arNone),
  (de:'Ȟ'; re:'Ȟ'; join:arNone),
  (de:'Ḥ'; re:'Ḥ'; join:arNone),
  (de:'Ḩ'; re:'Ḩ'; join:arNone),
  (de:'Ḫ'; re:'Ḫ'; join:arNone),
  (de:'Ì'; re:'Ì'; join:arNone),
  (de:'Í'; re:'Í'; join:arNone),
  (de:'Î'; re:'Î'; join:arNone),
  (de:'Ĩ'; re:'Ĩ'; join:arNone),
  (de:'Ī'; re:'Ī'; join:arNone),
  (de:'Ĭ'; re:'Ĭ'; join:arNone),
  (de:'İ'; re:'İ'; join:arNone),
  (de:'Ï'; re:'Ï'; join:arNone),
  (de:'Ỉ'; re:'Ỉ'; join:arNone),
  (de:'Ǐ'; re:'Ǐ'; join:arNone),
  (de:'Ȉ'; re:'Ȉ'; join:arNone),
  (de:'Ȋ'; re:'Ȋ'; join:arNone),
  (de:'Ị'; re:'Ị'; join:arNone),
  (de:'Į'; re:'Į'; join:arNone),
  (de:'Ḭ'; re:'Ḭ'; join:arNone),
  (de:'Ĵ'; re:'Ĵ'; join:arNone),
  (de:'Ḱ'; re:'Ḱ'; join:arNone),
  (de:'Ǩ'; re:'Ǩ'; join:arNone),
  (de:'Ḳ'; re:'Ḳ'; join:arNone),
  (de:'Ķ'; re:'Ķ'; join:arNone),
  (de:'Ḵ'; re:'Ḵ'; join:arNone),
  (de:'Ĺ'; re:'Ĺ'; join:arNone),
  (de:'Ľ'; re:'Ľ'; join:arNone),
  (de:'Ḷ'; re:'Ḷ'; join:arNone),
  (de:'Ļ'; re:'Ļ'; join:arNone),
  (de:'Ḽ'; re:'Ḽ'; join:arNone),
  (de:'Ḻ'; re:'Ḻ'; join:arNone),
  (de:'Ḿ'; re:'Ḿ'; join:arNone),
  (de:'Ṁ'; re:'Ṁ'; join:arNone),
  (de:'Ṃ'; re:'Ṃ'; join:arNone),
  (de:'Ǹ'; re:'Ǹ'; join:arNone),
  (de:'Ń'; re:'Ń'; join:arNone),
  (de:'Ñ'; re:'Ñ'; join:arNone),
  (de:'Ṅ'; re:'Ṅ'; join:arNone),
  (de:'Ň'; re:'Ň'; join:arNone),
  (de:'Ṇ'; re:'Ṇ'; join:arNone),
  (de:'Ņ'; re:'Ņ'; join:arNone),
  (de:'Ṋ'; re:'Ṋ'; join:arNone),
  (de:'Ṉ'; re:'Ṉ'; join:arNone),
  (de:'Ò'; re:'Ò'; join:arNone),
  (de:'Ó'; re:'Ó'; join:arNone),
  (de:'Ô'; re:'Ô'; join:arNone),
  (de:'Õ'; re:'Õ'; join:arNone),
  (de:'Ō'; re:'Ō'; join:arNone),
  (de:'Ŏ'; re:'Ŏ'; join:arNone),
  (de:'Ȯ'; re:'Ȯ'; join:arNone),
  (de:'Ö'; re:'Ö'; join:arNone),
  (de:'Ỏ'; re:'Ỏ'; join:arNone),
  (de:'Ő'; re:'Ő'; join:arNone),
  (de:'Ǒ'; re:'Ǒ'; join:arNone),
  (de:'Ȍ'; re:'Ȍ'; join:arNone),
  (de:'Ȏ'; re:'Ȏ'; join:arNone),
  (de:'Ơ'; re:'Ơ'; join:arNone),
  (de:'Ọ'; re:'Ọ'; join:arNone),
  (de:'Ǫ'; re:'Ǫ'; join:arNone),
  (de:'Ṕ'; re:'Ṕ'; join:arNone),
  (de:'Ṗ'; re:'Ṗ'; join:arNone),
  (de:'Ŕ'; re:'Ŕ'; join:arNone),
  (de:'Ṙ'; re:'Ṙ'; join:arNone),
  (de:'Ř'; re:'Ř'; join:arNone),
  (de:'Ȑ'; re:'Ȑ'; join:arNone),
  (de:'Ȓ'; re:'Ȓ'; join:arNone),
  (de:'Ṛ'; re:'Ṛ'; join:arNone),
  (de:'Ŗ'; re:'Ŗ'; join:arNone),
  (de:'Ṟ'; re:'Ṟ'; join:arNone),
  (de:'Ś'; re:'Ś'; join:arNone),
  (de:'Ŝ'; re:'Ŝ'; join:arNone),
  (de:'Ṡ'; re:'Ṡ'; join:arNone),
  (de:'Š'; re:'Š'; join:arNone),
  (de:'Ṣ'; re:'Ṣ'; join:arNone),
  (de:'Ș'; re:'Ș'; join:arNone),
  (de:'Ş'; re:'Ş'; join:arNone),
  (de:'Ṫ'; re:'Ṫ'; join:arNone),
  (de:'Ť'; re:'Ť'; join:arNone),
  (de:'Ṭ'; re:'Ṭ'; join:arNone),
  (de:'Ț'; re:'Ț'; join:arNone),
  (de:'Ţ'; re:'Ţ'; join:arNone),
  (de:'Ṱ'; re:'Ṱ'; join:arNone),
  (de:'Ṯ'; re:'Ṯ'; join:arNone),
  (de:'Ù'; re:'Ù'; join:arNone),
  (de:'Ú'; re:'Ú'; join:arNone),
  (de:'Û'; re:'Û'; join:arNone),
  (de:'Ũ'; re:'Ũ'; join:arNone),
  (de:'Ū'; re:'Ū'; join:arNone),
  (de:'Ŭ'; re:'Ŭ'; join:arNone),
  (de:'Ü'; re:'Ü'; join:arNone),
  (de:'Ủ'; re:'Ủ'; join:arNone),
  (de:'Ů'; re:'Ů'; join:arNone),
  (de:'Ű'; re:'Ű'; join:arNone),
  (de:'Ǔ'; re:'Ǔ'; join:arNone),
  (de:'Ȕ'; re:'Ȕ'; join:arNone),
  (de:'Ȗ'; re:'Ȗ'; join:arNone),
  (de:'Ư'; re:'Ư'; join:arNone),
  (de:'Ụ'; re:'Ụ'; join:arNone),
  (de:'Ṳ'; re:'Ṳ'; join:arNone),
  (de:'Ų'; re:'Ų'; join:arNone),
  (de:'Ṷ'; re:'Ṷ'; join:arNone),
  (de:'Ṵ'; re:'Ṵ'; join:arNone),
  (de:'Ṽ'; re:'Ṽ'; join:arNone),
  (de:'Ṿ'; re:'Ṿ'; join:arNone),
  (de:'Ẁ'; re:'Ẁ'; join:arNone),
  (de:'Ẃ'; re:'Ẃ'; join:arNone),
  (de:'Ŵ'; re:'Ŵ'; join:arNone),
  (de:'Ẇ'; re:'Ẇ'; join:arNone),
  (de:'Ẅ'; re:'Ẅ'; join:arNone),
  (de:'Ẉ'; re:'Ẉ'; join:arNone),
  (de:'Ẋ'; re:'Ẋ'; join:arNone),
  (de:'Ẍ'; re:'Ẍ'; join:arNone),
  (de:'Ỳ'; re:'Ỳ'; join:arNone),
  (de:'Ý'; re:'Ý'; join:arNone),
  (de:'Ŷ'; re:'Ŷ'; join:arNone),
  (de:'Ỹ'; re:'Ỹ'; join:arNone),
  (de:'Ȳ'; re:'Ȳ'; join:arNone),
  (de:'Ẏ'; re:'Ẏ'; join:arNone),
  (de:'Ÿ'; re:'Ÿ'; join:arNone),
  (de:'Ỷ'; re:'Ỷ'; join:arNone),
  (de:'Ỵ'; re:'Ỵ'; join:arNone),
  (de:'Ź'; re:'Ź'; join:arNone),
  (de:'Ẑ'; re:'Ẑ'; join:arNone),
  (de:'Ż'; re:'Ż'; join:arNone),
  (de:'Ž'; re:'Ž'; join:arNone),
  (de:'Ẓ'; re:'Ẓ'; join:arNone),
  (de:'Ẕ'; re:'Ẕ'; join:arNone),
  (de:'à'; re:'à'; join:arNone),
  (de:'á'; re:'á'; join:arNone),
  (de:'â'; re:'â'; join:arNone),
  (de:'ã'; re:'ã'; join:arNone),
  (de:'ā'; re:'ā'; join:arNone),
  (de:'ă'; re:'ă'; join:arNone),
  (de:'ȧ'; re:'ȧ'; join:arNone),
  (de:'ä'; re:'ä'; join:arNone),
  (de:'ả'; re:'ả'; join:arNone),
  (de:'å'; re:'å'; join:arNone),
  (de:'ǎ'; re:'ǎ'; join:arNone),
  (de:'ȁ'; re:'ȁ'; join:arNone),
  (de:'ȃ'; re:'ȃ'; join:arNone),
  (de:'ạ'; re:'ạ'; join:arNone),
  (de:'ḁ'; re:'ḁ'; join:arNone),
  (de:'ą'; re:'ą'; join:arNone),
  (de:'ḃ'; re:'ḃ'; join:arNone),
  (de:'ḅ'; re:'ḅ'; join:arNone),
  (de:'ḇ'; re:'ḇ'; join:arNone),
  (de:'ć'; re:'ć'; join:arNone),
  (de:'ĉ'; re:'ĉ'; join:arNone),
  (de:'ċ'; re:'ċ'; join:arNone),
  (de:'č'; re:'č'; join:arNone),
  (de:'ç'; re:'ç'; join:arNone),
  (de:'ḋ'; re:'ḋ'; join:arNone),
  (de:'ď'; re:'ď'; join:arNone),
  (de:'ḍ'; re:'ḍ'; join:arNone),
  (de:'ḑ'; re:'ḑ'; join:arNone),
  (de:'ḓ'; re:'ḓ'; join:arNone),
  (de:'ḏ'; re:'ḏ'; join:arNone),
  (de:'è'; re:'è'; join:arNone),
  (de:'é'; re:'é'; join:arNone),
  (de:'ê'; re:'ê'; join:arNone),
  (de:'ẽ'; re:'ẽ'; join:arNone),
  (de:'ē'; re:'ē'; join:arNone),
  (de:'ĕ'; re:'ĕ'; join:arNone),
  (de:'ė'; re:'ė'; join:arNone),
  (de:'ë'; re:'ë'; join:arNone),
  (de:'ẻ'; re:'ẻ'; join:arNone),
  (de:'ě'; re:'ě'; join:arNone),
  (de:'ȅ'; re:'ȅ'; join:arNone),
  (de:'ȇ'; re:'ȇ'; join:arNone),
  (de:'ẹ'; re:'ẹ'; join:arNone),
  (de:'ȩ'; re:'ȩ'; join:arNone),
  (de:'ę'; re:'ę'; join:arNone),
  (de:'ḙ'; re:'ḙ'; join:arNone),
  (de:'ḛ'; re:'ḛ'; join:arNone),
  (de:'ff'; re:'ﬀ'; join:arNone),
  (de:'ffi'; re:'ﬃ'; join:arNone),
  (de:'ffl'; re:'ﬄ'; join:arNone),
  (de:'fi'; re:'ﬁ'; join:arNone),
  (de:'fl'; re:'ﬂ'; join:arNone),
  (de:'ḟ'; re:'ḟ'; join:arNone),
  (de:'ǵ'; re:'ǵ'; join:arNone),
  (de:'ĝ'; re:'ĝ'; join:arNone),
  (de:'ḡ'; re:'ḡ'; join:arNone),
  (de:'ğ'; re:'ğ'; join:arNone),
  (de:'ġ'; re:'ġ'; join:arNone),
  (de:'ǧ'; re:'ǧ'; join:arNone),
  (de:'ģ'; re:'ģ'; join:arNone),
  (de:'ĥ'; re:'ĥ'; join:arNone),
  (de:'ḣ'; re:'ḣ'; join:arNone),
  (de:'ḧ'; re:'ḧ'; join:arNone),
  (de:'ȟ'; re:'ȟ'; join:arNone),
  (de:'ḥ'; re:'ḥ'; join:arNone),
  (de:'ḩ'; re:'ḩ'; join:arNone),
  (de:'ḫ'; re:'ḫ'; join:arNone),
  (de:'ẖ'; re:'ẖ'; join:arNone),
  (de:'ì'; re:'ì'; join:arNone),
  (de:'í'; re:'í'; join:arNone),
  (de:'î'; re:'î'; join:arNone),
  (de:'ĩ'; re:'ĩ'; join:arNone),
  (de:'ī'; re:'ī'; join:arNone),
  (de:'ĭ'; re:'ĭ'; join:arNone),
  (de:'ï'; re:'ï'; join:arNone),
  (de:'ỉ'; re:'ỉ'; join:arNone),
  (de:'ǐ'; re:'ǐ'; join:arNone),
  (de:'ȉ'; re:'ȉ'; join:arNone),
  (de:'ȋ'; re:'ȋ'; join:arNone),
  (de:'ị'; re:'ị'; join:arNone),
  (de:'į'; re:'į'; join:arNone),
  (de:'ḭ'; re:'ḭ'; join:arNone),
  (de:'ĵ'; re:'ĵ'; join:arNone),
  (de:'ǰ'; re:'ǰ'; join:arNone),
  (de:'ḱ'; re:'ḱ'; join:arNone),
  (de:'ǩ'; re:'ǩ'; join:arNone),
  (de:'ḳ'; re:'ḳ'; join:arNone),
  (de:'ķ'; re:'ķ'; join:arNone),
  (de:'ḵ'; re:'ḵ'; join:arNone),
  (de:'ĺ'; re:'ĺ'; join:arNone),
  (de:'ľ'; re:'ľ'; join:arNone),
  (de:'ḷ'; re:'ḷ'; join:arNone),
  (de:'ļ'; re:'ļ'; join:arNone),
  (de:'ḽ'; re:'ḽ'; join:arNone),
  (de:'ḻ'; re:'ḻ'; join:arNone),
  (de:'ḿ'; re:'ḿ'; join:arNone),
  (de:'ṁ'; re:'ṁ'; join:arNone),
  (de:'ṃ'; re:'ṃ'; join:arNone),
  (de:'ǹ'; re:'ǹ'; join:arNone),
  (de:'ń'; re:'ń'; join:arNone),
  (de:'ñ'; re:'ñ'; join:arNone),
  (de:'ṅ'; re:'ṅ'; join:arNone),
  (de:'ň'; re:'ň'; join:arNone),
  (de:'ṇ'; re:'ṇ'; join:arNone),
  (de:'ņ'; re:'ņ'; join:arNone),
  (de:'ṋ'; re:'ṋ'; join:arNone),
  (de:'ṉ'; re:'ṉ'; join:arNone),
  (de:'ò'; re:'ò'; join:arNone),
  (de:'ó'; re:'ó'; join:arNone),
  (de:'ô'; re:'ô'; join:arNone),
  (de:'õ'; re:'õ'; join:arNone),
  (de:'ō'; re:'ō'; join:arNone),
  (de:'ŏ'; re:'ŏ'; join:arNone),
  (de:'ȯ'; re:'ȯ'; join:arNone),
  (de:'ö'; re:'ö'; join:arNone),
  (de:'ỏ'; re:'ỏ'; join:arNone),
  (de:'ő'; re:'ő'; join:arNone),
  (de:'ǒ'; re:'ǒ'; join:arNone),
  (de:'ȍ'; re:'ȍ'; join:arNone),
  (de:'ȏ'; re:'ȏ'; join:arNone),
  (de:'ơ'; re:'ơ'; join:arNone),
  (de:'ọ'; re:'ọ'; join:arNone),
  (de:'ǫ'; re:'ǫ'; join:arNone),
  (de:'ṕ'; re:'ṕ'; join:arNone),
  (de:'ṗ'; re:'ṗ'; join:arNone),
  (de:'ŕ'; re:'ŕ'; join:arNone),
  (de:'ṙ'; re:'ṙ'; join:arNone),
  (de:'ř'; re:'ř'; join:arNone),
  (de:'ȑ'; re:'ȑ'; join:arNone),
  (de:'ȓ'; re:'ȓ'; join:arNone),
  (de:'ṛ'; re:'ṛ'; join:arNone),
  (de:'ŗ'; re:'ŗ'; join:arNone),
  (de:'ṟ'; re:'ṟ'; join:arNone),
  (de:'ś'; re:'ś'; join:arNone),
  (de:'ŝ'; re:'ŝ'; join:arNone),
  (de:'ṡ'; re:'ṡ'; join:arNone),
  (de:'š'; re:'š'; join:arNone),
  (de:'ṣ'; re:'ṣ'; join:arNone),
  (de:'ș'; re:'ș'; join:arNone),
  (de:'ş'; re:'ş'; join:arNone),
  (de:'ṫ'; re:'ṫ'; join:arNone),
  (de:'ẗ'; re:'ẗ'; join:arNone),
  (de:'ť'; re:'ť'; join:arNone),
  (de:'ṭ'; re:'ṭ'; join:arNone),
  (de:'ț'; re:'ț'; join:arNone),
  (de:'ţ'; re:'ţ'; join:arNone),
  (de:'ṱ'; re:'ṱ'; join:arNone),
  (de:'ṯ'; re:'ṯ'; join:arNone),
  (de:'ù'; re:'ù'; join:arNone),
  (de:'ú'; re:'ú'; join:arNone),
  (de:'û'; re:'û'; join:arNone),
  (de:'ũ'; re:'ũ'; join:arNone),
  (de:'ū'; re:'ū'; join:arNone),
  (de:'ŭ'; re:'ŭ'; join:arNone),
  (de:'ü'; re:'ü'; join:arNone),
  (de:'ủ'; re:'ủ'; join:arNone),
  (de:'ů'; re:'ů'; join:arNone),
  (de:'ű'; re:'ű'; join:arNone),
  (de:'ǔ'; re:'ǔ'; join:arNone),
  (de:'ȕ'; re:'ȕ'; join:arNone),
  (de:'ȗ'; re:'ȗ'; join:arNone),
  (de:'ư'; re:'ư'; join:arNone),
  (de:'ụ'; re:'ụ'; join:arNone),
  (de:'ṳ'; re:'ṳ'; join:arNone),
  (de:'ų'; re:'ų'; join:arNone),
  (de:'ṷ'; re:'ṷ'; join:arNone),
  (de:'ṵ'; re:'ṵ'; join:arNone),
  (de:'ṽ'; re:'ṽ'; join:arNone),
  (de:'ṿ'; re:'ṿ'; join:arNone),
  (de:'ẁ'; re:'ẁ'; join:arNone),
  (de:'ẃ'; re:'ẃ'; join:arNone),
  (de:'ŵ'; re:'ŵ'; join:arNone),
  (de:'ẇ'; re:'ẇ'; join:arNone),
  (de:'ẅ'; re:'ẅ'; join:arNone),
  (de:'ẘ'; re:'ẘ'; join:arNone),
  (de:'ẉ'; re:'ẉ'; join:arNone),
  (de:'ẋ'; re:'ẋ'; join:arNone),
  (de:'ẍ'; re:'ẍ'; join:arNone),
  (de:'ỳ'; re:'ỳ'; join:arNone),
  (de:'ý'; re:'ý'; join:arNone),
  (de:'ŷ'; re:'ŷ'; join:arNone),
  (de:'ỹ'; re:'ỹ'; join:arNone),
  (de:'ȳ'; re:'ȳ'; join:arNone),
  (de:'ẏ'; re:'ẏ'; join:arNone),
  (de:'ÿ'; re:'ÿ'; join:arNone),
  (de:'ỷ'; re:'ỷ'; join:arNone),
  (de:'ẙ'; re:'ẙ'; join:arNone),
  (de:'ỵ'; re:'ỵ'; join:arNone),
  (de:'ź'; re:'ź'; join:arNone),
  (de:'ẑ'; re:'ẑ'; join:arNone),
  (de:'ż'; re:'ż'; join:arNone),
  (de:'ž'; re:'ž'; join:arNone),
  (de:'ẓ'; re:'ẓ'; join:arNone),
  (de:'ẕ'; re:'ẕ'; join:arNone),
  (de:'῭'; re:'῭'; join:arNone),
  (de:'΅'; re:'΅'; join:arNone),
  (de:'῁'; re:'῁'; join:arNone),
  (de:'Ầ'; re:'Ầ'; join:arNone),
  (de:'Ấ'; re:'Ấ'; join:arNone),
  (de:'Ẫ'; re:'Ẫ'; join:arNone),
  (de:'Ẩ'; re:'Ẩ'; join:arNone),
  (de:'Ǟ'; re:'Ǟ'; join:arNone),
  (de:'Ǻ'; re:'Ǻ'; join:arNone),
  (de:'Ǽ'; re:'Ǽ'; join:arNone),
  (de:'Ǣ'; re:'Ǣ'; join:arNone),
  (de:'Ḉ'; re:'Ḉ'; join:arNone),
  (de:'Ề'; re:'Ề'; join:arNone),
  (de:'Ế'; re:'Ế'; join:arNone),
  (de:'Ễ'; re:'Ễ'; join:arNone),
  (de:'Ể'; re:'Ể'; join:arNone),
  (de:'Ḯ'; re:'Ḯ'; join:arNone),
  (de:'Ồ'; re:'Ồ'; join:arNone),
  (de:'Ố'; re:'Ố'; join:arNone),
  (de:'Ỗ'; re:'Ỗ'; join:arNone),
  (de:'Ổ'; re:'Ổ'; join:arNone),
  (de:'Ṍ'; re:'Ṍ'; join:arNone),
  (de:'Ȭ'; re:'Ȭ'; join:arNone),
  (de:'Ṏ'; re:'Ṏ'; join:arNone),
  (de:'Ȫ'; re:'Ȫ'; join:arNone),
  (de:'Ǿ'; re:'Ǿ'; join:arNone),
  (de:'Ǜ'; re:'Ǜ'; join:arNone),
  (de:'Ǘ'; re:'Ǘ'; join:arNone),
  (de:'Ǖ'; re:'Ǖ'; join:arNone),
  (de:'Ǚ'; re:'Ǚ'; join:arNone),
  (de:'ầ'; re:'ầ'; join:arNone),
  (de:'ấ'; re:'ấ'; join:arNone),
  (de:'ẫ'; re:'ẫ'; join:arNone),
  (de:'ẩ'; re:'ẩ'; join:arNone),
  (de:'ǟ'; re:'ǟ'; join:arNone),
  (de:'ǻ'; re:'ǻ'; join:arNone),
  (de:'ǽ'; re:'ǽ'; join:arNone),
  (de:'ǣ'; re:'ǣ'; join:arNone),
  (de:'ḉ'; re:'ḉ'; join:arNone),
  (de:'ề'; re:'ề'; join:arNone),
  (de:'ế'; re:'ế'; join:arNone),
  (de:'ễ'; re:'ễ'; join:arNone),
  (de:'ể'; re:'ể'; join:arNone),
  (de:'ḯ'; re:'ḯ'; join:arNone),
  (de:'ồ'; re:'ồ'; join:arNone),
  (de:'ố'; re:'ố'; join:arNone),
  (de:'ỗ'; re:'ỗ'; join:arNone),
  (de:'ổ'; re:'ổ'; join:arNone),
  (de:'ṍ'; re:'ṍ'; join:arNone),
  (de:'ȭ'; re:'ȭ'; join:arNone),
  (de:'ṏ'; re:'ṏ'; join:arNone),
  (de:'ȫ'; re:'ȫ'; join:arNone),
  (de:'ǿ'; re:'ǿ'; join:arNone),
  (de:'ǜ'; re:'ǜ'; join:arNone),
  (de:'ǘ'; re:'ǘ'; join:arNone),
  (de:'ǖ'; re:'ǖ'; join:arNone),
  (de:'ǚ'; re:'ǚ'; join:arNone),
  (de:'Ằ'; re:'Ằ'; join:arNone),
  (de:'Ắ'; re:'Ắ'; join:arNone),
  (de:'Ẵ'; re:'Ẵ'; join:arNone),
  (de:'Ẳ'; re:'Ẳ'; join:arNone),
  (de:'ằ'; re:'ằ'; join:arNone),
  (de:'ắ'; re:'ắ'; join:arNone),
  (de:'ẵ'; re:'ẵ'; join:arNone),
  (de:'ẳ'; re:'ẳ'; join:arNone),
  (de:'Ḕ'; re:'Ḕ'; join:arNone),
  (de:'Ḗ'; re:'Ḗ'; join:arNone),
  (de:'ḕ'; re:'ḕ'; join:arNone),
  (de:'ḗ'; re:'ḗ'; join:arNone),
  (de:'Ṑ'; re:'Ṑ'; join:arNone),
  (de:'Ṓ'; re:'Ṓ'; join:arNone),
  (de:'ṑ'; re:'ṑ'; join:arNone),
  (de:'ṓ'; re:'ṓ'; join:arNone),
  (de:'Ṥ'; re:'Ṥ'; join:arNone),
  (de:'ṥ'; re:'ṥ'; join:arNone),
  (de:'Ṧ'; re:'Ṧ'; join:arNone),
  (de:'ṧ'; re:'ṧ'; join:arNone),
  (de:'Ṹ'; re:'Ṹ'; join:arNone),
  (de:'ṹ'; re:'ṹ'; join:arNone),
  (de:'Ṻ'; re:'Ṻ'; join:arNone),
  (de:'ṻ'; re:'ṻ'; join:arNone),
  (de:'ẛ'; re:'ẛ'; join:arNone),
  (de:'Ờ'; re:'Ờ'; join:arNone),
  (de:'Ớ'; re:'Ớ'; join:arNone),
  (de:'Ỡ'; re:'Ỡ'; join:arNone),
  (de:'Ở'; re:'Ở'; join:arNone),
  (de:'Ợ'; re:'Ợ'; join:arNone),
  (de:'ờ'; re:'ờ'; join:arNone),
  (de:'ớ'; re:'ớ'; join:arNone),
  (de:'ỡ'; re:'ỡ'; join:arNone),
  (de:'ở'; re:'ở'; join:arNone),
  (de:'ợ'; re:'ợ'; join:arNone),
  (de:'Ừ'; re:'Ừ'; join:arNone),
  (de:'Ứ'; re:'Ứ'; join:arNone),
  (de:'Ữ'; re:'Ữ'; join:arNone),
  (de:'Ử'; re:'Ử'; join:arNone),
  (de:'Ự'; re:'Ự'; join:arNone),
  (de:'ừ'; re:'ừ'; join:arNone),
  (de:'ứ'; re:'ứ'; join:arNone),
  (de:'ữ'; re:'ữ'; join:arNone),
  (de:'ử'; re:'ử'; join:arNone),
  (de:'ự'; re:'ự'; join:arNone),
  (de:'Ǯ'; re:'Ǯ'; join:arNone),
  (de:'Ǭ'; re:'Ǭ'; join:arNone),
  (de:'ǭ'; re:'ǭ'; join:arNone),
  (de:'Ǡ'; re:'Ǡ'; join:arNone),
  (de:'ǡ'; re:'ǡ'; join:arNone),
  (de:'Ḝ'; re:'Ḝ'; join:arNone),
  (de:'ḝ'; re:'ḝ'; join:arNone),
  (de:'Ȱ'; re:'Ȱ'; join:arNone),
  (de:'ȱ'; re:'ȱ'; join:arNone),
  (de:'ǯ'; re:'ǯ'; join:arNone),
  (de:'Ὰ'; re:'Ὰ'; join:arNone),
  (de:'Ά'; re:'Ά'; join:arNone),
  (de:'Ᾱ'; re:'Ᾱ'; join:arNone),
  (de:'Ᾰ'; re:'Ᾰ'; join:arNone),
  (de:'Ἀ'; re:'Ἀ'; join:arNone),
  (de:'Ἁ'; re:'Ἁ'; join:arNone),
  (de:'ᾼ'; re:'ᾼ'; join:arNone),
  (de:'Ὲ'; re:'Ὲ'; join:arNone),
  (de:'Έ'; re:'Έ'; join:arNone),
  (de:'Ἐ'; re:'Ἐ'; join:arNone),
  (de:'Ἑ'; re:'Ἑ'; join:arNone),
  (de:'Ὴ'; re:'Ὴ'; join:arNone),
  (de:'Ή'; re:'Ή'; join:arNone),
  (de:'Ἠ'; re:'Ἠ'; join:arNone),
  (de:'Ἡ'; re:'Ἡ'; join:arNone),
  (de:'ῌ'; re:'ῌ'; join:arNone),
  (de:'Ὶ'; re:'Ὶ'; join:arNone),
  (de:'Ί'; re:'Ί'; join:arNone),
  (de:'Ῑ'; re:'Ῑ'; join:arNone),
  (de:'Ῐ'; re:'Ῐ'; join:arNone),
  (de:'Ϊ'; re:'Ϊ'; join:arNone),
  (de:'Ἰ'; re:'Ἰ'; join:arNone),
  (de:'Ἱ'; re:'Ἱ'; join:arNone),
  (de:'Ὸ'; re:'Ὸ'; join:arNone),
  (de:'Ό'; re:'Ό'; join:arNone),
  (de:'Ὀ'; re:'Ὀ'; join:arNone),
  (de:'Ὁ'; re:'Ὁ'; join:arNone),
  (de:'Ῥ'; re:'Ῥ'; join:arNone),
  (de:'Ὺ'; re:'Ὺ'; join:arNone),
  (de:'Ύ'; re:'Ύ'; join:arNone),
  (de:'Ῡ'; re:'Ῡ'; join:arNone),
  (de:'Ῠ'; re:'Ῠ'; join:arNone),
  (de:'Ϋ'; re:'Ϋ'; join:arNone),
  (de:'Ὑ'; re:'Ὑ'; join:arNone),
  (de:'Ὼ'; re:'Ὼ'; join:arNone),
  (de:'Ώ'; re:'Ώ'; join:arNone),
  (de:'Ὠ'; re:'Ὠ'; join:arNone),
  (de:'Ὡ'; re:'Ὡ'; join:arNone),
  (de:'ῼ'; re:'ῼ'; join:arNone),
  (de:'ᾴ'; re:'ᾴ'; join:arNone),
  (de:'ῄ'; re:'ῄ'; join:arNone),
  (de:'ὰ'; re:'ὰ'; join:arNone),
  (de:'ά'; re:'ά'; join:arNone),
  (de:'ᾱ'; re:'ᾱ'; join:arNone),
  (de:'ᾰ'; re:'ᾰ'; join:arNone),
  (de:'ἀ'; re:'ἀ'; join:arNone),
  (de:'ἁ'; re:'ἁ'; join:arNone),
  (de:'ᾶ'; re:'ᾶ'; join:arNone),
  (de:'ᾳ'; re:'ᾳ'; join:arNone),
  (de:'ὲ'; re:'ὲ'; join:arNone),
  (de:'έ'; re:'έ'; join:arNone),
  (de:'ἐ'; re:'ἐ'; join:arNone),
  (de:'ἑ'; re:'ἑ'; join:arNone),
  (de:'ὴ'; re:'ὴ'; join:arNone),
  (de:'ή'; re:'ή'; join:arNone),
  (de:'ἠ'; re:'ἠ'; join:arNone),
  (de:'ἡ'; re:'ἡ'; join:arNone),
  (de:'ῆ'; re:'ῆ'; join:arNone),
  (de:'ῃ'; re:'ῃ'; join:arNone),
  (de:'ὶ'; re:'ὶ'; join:arNone),
  (de:'ί'; re:'ί'; join:arNone),
  (de:'ῑ'; re:'ῑ'; join:arNone),
  (de:'ῐ'; re:'ῐ'; join:arNone),
  (de:'ϊ'; re:'ϊ'; join:arNone),
  (de:'ἰ'; re:'ἰ'; join:arNone),
  (de:'ἱ'; re:'ἱ'; join:arNone),
  (de:'ῖ'; re:'ῖ'; join:arNone),
  (de:'ὸ'; re:'ὸ'; join:arNone),
  (de:'ό'; re:'ό'; join:arNone),
  (de:'ὀ'; re:'ὀ'; join:arNone),
  (de:'ὁ'; re:'ὁ'; join:arNone),
  (de:'ῤ'; re:'ῤ'; join:arNone),
  (de:'ῥ'; re:'ῥ'; join:arNone),
  (de:'ὺ'; re:'ὺ'; join:arNone),
  (de:'ύ'; re:'ύ'; join:arNone),
  (de:'ῡ'; re:'ῡ'; join:arNone),
  (de:'ῠ'; re:'ῠ'; join:arNone),
  (de:'ϋ'; re:'ϋ'; join:arNone),
  (de:'ὐ'; re:'ὐ'; join:arNone),
  (de:'ὑ'; re:'ὑ'; join:arNone),
  (de:'ῦ'; re:'ῦ'; join:arNone),
  (de:'ὼ'; re:'ὼ'; join:arNone),
  (de:'ώ'; re:'ώ'; join:arNone),
  (de:'ὠ'; re:'ὠ'; join:arNone),
  (de:'ὡ'; re:'ὡ'; join:arNone),
  (de:'ῶ'; re:'ῶ'; join:arNone),
  (de:'ῳ'; re:'ῳ'; join:arNone),
  (de:'ῒ'; re:'ῒ'; join:arNone),
  (de:'ΐ'; re:'ΐ'; join:arNone),
  (de:'ῗ'; re:'ῗ'; join:arNone),
  (de:'ῢ'; re:'ῢ'; join:arNone),
  (de:'ΰ'; re:'ΰ'; join:arNone),
  (de:'ῧ'; re:'ῧ'; join:arNone),
  (de:'ῴ'; re:'ῴ'; join:arNone),
  (de:'ϓ'; re:'ϓ'; join:arNone),
  (de:'ϔ'; re:'ϔ'; join:arNone),
  (de:'Ї'; re:'Ї'; join:arNone),
  (de:'Ӑ'; re:'Ӑ'; join:arNone),
  (de:'Ӓ'; re:'Ӓ'; join:arNone),
  (de:'Ѓ'; re:'Ѓ'; join:arNone),
  (de:'Ѐ'; re:'Ѐ'; join:arNone),
  (de:'Ӗ'; re:'Ӗ'; join:arNone),
  (de:'Ё'; re:'Ё'; join:arNone),
  (de:'Ӂ'; re:'Ӂ'; join:arNone),
  (de:'Ӝ'; re:'Ӝ'; join:arNone),
  (de:'Ӟ'; re:'Ӟ'; join:arNone),
  (de:'Ѝ'; re:'Ѝ'; join:arNone),
  (de:'Ӣ'; re:'Ӣ'; join:arNone),
  (de:'Й'; re:'Й'; join:arNone),
  (de:'Ӥ'; re:'Ӥ'; join:arNone),
  (de:'Ќ'; re:'Ќ'; join:arNone),
  (de:'Ӧ'; re:'Ӧ'; join:arNone),
  (de:'Ӯ'; re:'Ӯ'; join:arNone),
  (de:'Ў'; re:'Ў'; join:arNone),
  (de:'Ӱ'; re:'Ӱ'; join:arNone),
  (de:'Ӳ'; re:'Ӳ'; join:arNone),
  (de:'Ӵ'; re:'Ӵ'; join:arNone),
  (de:'Ӹ'; re:'Ӹ'; join:arNone),
  (de:'Ӭ'; re:'Ӭ'; join:arNone),
  (de:'ӑ'; re:'ӑ'; join:arNone),
  (de:'ӓ'; re:'ӓ'; join:arNone),
  (de:'ѓ'; re:'ѓ'; join:arNone),
  (de:'ѐ'; re:'ѐ'; join:arNone),
  (de:'ӗ'; re:'ӗ'; join:arNone),
  (de:'ё'; re:'ё'; join:arNone),
  (de:'ӂ'; re:'ӂ'; join:arNone),
  (de:'ӝ'; re:'ӝ'; join:arNone),
  (de:'ӟ'; re:'ӟ'; join:arNone),
  (de:'ѝ'; re:'ѝ'; join:arNone),
  (de:'ӣ'; re:'ӣ'; join:arNone),
  (de:'й'; re:'й'; join:arNone),
  (de:'ӥ'; re:'ӥ'; join:arNone),
  (de:'ќ'; re:'ќ'; join:arNone),
  (de:'ӧ'; re:'ӧ'; join:arNone),
  (de:'ӯ'; re:'ӯ'; join:arNone),
  (de:'ў'; re:'ў'; join:arNone),
  (de:'ӱ'; re:'ӱ'; join:arNone),
  (de:'ӳ'; re:'ӳ'; join:arNone),
  (de:'ӵ'; re:'ӵ'; join:arNone),
  (de:'ӹ'; re:'ӹ'; join:arNone),
  (de:'ӭ'; re:'ӭ'; join:arNone),
  (de:'ї'; re:'ї'; join:arNone),
  (de:'Ѷ'; re:'Ѷ'; join:arNone),
  (de:'ѷ'; re:'ѷ'; join:arNone),
  (de:'Ӛ'; re:'Ӛ'; join:arNone),
  (de:'ӛ'; re:'ӛ'; join:arNone),
  (de:'Ӫ'; re:'Ӫ'; join:arNone),
  (de:'ӫ'; re:'ӫ'; join:arNone),
  (de:'אַ'; re:'אַ'; join:arNone),
  (de:'אָ'; re:'אָ'; join:arNone),
  (de:'אּ'; re:'אּ'; join:arNone),
  (de:'בּ'; re:'בּ'; join:arNone),
  (de:'בֿ'; re:'בֿ'; join:arNone),
  (de:'גּ'; re:'גּ'; join:arNone),
  (de:'דּ'; re:'דּ'; join:arNone),
  (de:'הּ'; re:'הּ'; join:arNone),
  (de:'וֹ'; re:'וֹ'; join:arNone),
  (de:'וּ'; re:'וּ'; join:arNone),
  (de:'זּ'; re:'זּ'; join:arNone),
  (de:'טּ'; re:'טּ'; join:arNone),
  (de:'יִ'; re:'יִ'; join:arNone),
  (de:'יּ'; re:'יּ'; join:arNone),
  (de:'ךּ'; re:'ךּ'; join:arNone),
  (de:'כּ'; re:'כּ'; join:arNone),
  (de:'כֿ'; re:'כֿ'; join:arNone),
  (de:'לּ'; re:'לּ'; join:arNone),
  (de:'מּ'; re:'מּ'; join:arNone),
  (de:'נּ'; re:'נּ'; join:arNone),
  (de:'סּ'; re:'סּ'; join:arNone),
  (de:'ףּ'; re:'ףּ'; join:arNone),
  (de:'פּ'; re:'פּ'; join:arNone),
  (de:'פֿ'; re:'פֿ'; join:arNone),
  (de:'צּ'; re:'צּ'; join:arNone),
  (de:'קּ'; re:'קּ'; join:arNone),
  (de:'רּ'; re:'רּ'; join:arNone),
  (de:'שּ'; re:'שּ'; join:arNone),
  (de:'שׁ'; re:'שׁ'; join:arNone),
  (de:'שׂ'; re:'שׂ'; join:arNone),
  (de:'תּ'; re:'תּ'; join:arNone),
  (de:'ײַ'; re:'ײַ'; join:arNone),
  (de:'ء'; re:'ﺀ'; join:arIsolated),
  (de:'آ'; re:'ﺁ'; join:arIsolated),
  (de:'آ'; re:'ﺂ'; join:arFinal),
  (de:'أ'; re:'ﺃ'; join:arIsolated),
  (de:'أ'; re:'ﺄ'; join:arFinal),
  (de:'ؤ'; re:'ﺅ'; join:arIsolated),
  (de:'ؤ'; re:'ﺆ'; join:arFinal),
  (de:'إ'; re:'ﺇ'; join:arIsolated),
  (de:'إ'; re:'ﺈ'; join:arFinal),
  (de:'ئ'; re:'ﺉ'; join:arIsolated),
  (de:'ئ'; re:'ﺊ'; join:arFinal),
  (de:'ئ'; re:'ﺋ'; join:arInitial),
  (de:'ئ'; re:'ﺌ'; join:arMedial),
  (de:'ا'; re:'ﺍ'; join:arIsolated),
  (de:'ا'; re:'ﺎ'; join:arFinal),
  (de:'آ'; re:'آ'; join:arNone),
  (de:'أ'; re:'أ'; join:arNone),
  (de:'إ'; re:'إ'; join:arNone),
  (de:'ب'; re:'ﺏ'; join:arIsolated),
  (de:'ب'; re:'ﺐ'; join:arFinal),
  (de:'ب'; re:'ﺑ'; join:arInitial),
  (de:'ب'; re:'ﺒ'; join:arMedial),
  (de:'ة'; re:'ﺓ'; join:arIsolated),
  (de:'ة'; re:'ﺔ'; join:arFinal),
  (de:'ت'; re:'ﺕ'; join:arIsolated),
  (de:'ت'; re:'ﺖ'; join:arFinal),
  (de:'ت'; re:'ﺗ'; join:arInitial),
  (de:'ت'; re:'ﺘ'; join:arMedial),
  (de:'ث'; re:'ﺙ'; join:arIsolated),
  (de:'ث'; re:'ﺚ'; join:arFinal),
  (de:'ث'; re:'ﺛ'; join:arInitial),
  (de:'ث'; re:'ﺜ'; join:arMedial),
  (de:'ج'; re:'ﺝ'; join:arIsolated),
  (de:'ج'; re:'ﺞ'; join:arFinal),
  (de:'ج'; re:'ﺟ'; join:arInitial),
  (de:'ج'; re:'ﺠ'; join:arMedial),
  (de:'ح'; re:'ﺡ'; join:arIsolated),
  (de:'ح'; re:'ﺢ'; join:arFinal),
  (de:'ح'; re:'ﺣ'; join:arInitial),
  (de:'ح'; re:'ﺤ'; join:arMedial),
  (de:'خ'; re:'ﺥ'; join:arIsolated),
  (de:'خ'; re:'ﺦ'; join:arFinal),
  (de:'خ'; re:'ﺧ'; join:arInitial),
  (de:'خ'; re:'ﺨ'; join:arMedial),
  (de:'د'; re:'ﺩ'; join:arIsolated),
  (de:'د'; re:'ﺪ'; join:arFinal),
  (de:'ذ'; re:'ﺫ'; join:arIsolated),
  (de:'ذ'; re:'ﺬ'; join:arFinal),
  (de:'ر'; re:'ﺭ'; join:arIsolated),
  (de:'ر'; re:'ﺮ'; join:arFinal),
  (de:'ز'; re:'ﺯ'; join:arIsolated),
  (de:'ز'; re:'ﺰ'; join:arFinal),
  (de:'س'; re:'ﺱ'; join:arIsolated),
  (de:'س'; re:'ﺲ'; join:arFinal),
  (de:'س'; re:'ﺳ'; join:arInitial),
  (de:'س'; re:'ﺴ'; join:arMedial),
  (de:'ش'; re:'ﺵ'; join:arIsolated),
  (de:'ش'; re:'ﺶ'; join:arFinal),
  (de:'ش'; re:'ﺷ'; join:arInitial),
  (de:'ش'; re:'ﺸ'; join:arMedial),
  (de:'ص'; re:'ﺹ'; join:arIsolated),
  (de:'ص'; re:'ﺺ'; join:arFinal),
  (de:'ص'; re:'ﺻ'; join:arInitial),
  (de:'ص'; re:'ﺼ'; join:arMedial),
  (de:'ض'; re:'ﺽ'; join:arIsolated),
  (de:'ض'; re:'ﺾ'; join:arFinal),
  (de:'ض'; re:'ﺿ'; join:arInitial),
  (de:'ض'; re:'ﻀ'; join:arMedial),
  (de:'ط'; re:'ﻁ'; join:arIsolated),
  (de:'ط'; re:'ﻂ'; join:arFinal),
  (de:'ط'; re:'ﻃ'; join:arInitial),
  (de:'ط'; re:'ﻄ'; join:arMedial),
  (de:'ظ'; re:'ﻅ'; join:arIsolated),
  (de:'ظ'; re:'ﻆ'; join:arFinal),
  (de:'ظ'; re:'ﻇ'; join:arInitial),
  (de:'ظ'; re:'ﻈ'; join:arMedial),
  (de:'ع'; re:'ﻉ'; join:arIsolated),
  (de:'ع'; re:'ﻊ'; join:arFinal),
  (de:'ع'; re:'ﻋ'; join:arInitial),
  (de:'ع'; re:'ﻌ'; join:arMedial),
  (de:'غ'; re:'ﻍ'; join:arIsolated),
  (de:'غ'; re:'ﻎ'; join:arFinal),
  (de:'غ'; re:'ﻏ'; join:arInitial),
  (de:'غ'; re:'ﻐ'; join:arMedial),
  (de:'ف'; re:'ﻑ'; join:arIsolated),
  (de:'ف'; re:'ﻒ'; join:arFinal),
  (de:'ف'; re:'ﻓ'; join:arInitial),
  (de:'ف'; re:'ﻔ'; join:arMedial),
  (de:'ق'; re:'ﻕ'; join:arIsolated),
  (de:'ق'; re:'ﻖ'; join:arFinal),
  (de:'ق'; re:'ﻗ'; join:arInitial),
  (de:'ق'; re:'ﻘ'; join:arMedial),
  (de:'ك'; re:'ﻙ'; join:arIsolated),
  (de:'ك'; re:'ﻚ'; join:arFinal),
  (de:'ك'; re:'ﻛ'; join:arInitial),
  (de:'ك'; re:'ﻜ'; join:arMedial),
  (de:'ل'; re:'ﻝ'; join:arIsolated),
  (de:'ل'; re:'ﻞ'; join:arFinal),
  (de:'ل'; re:'ﻟ'; join:arInitial),
  (de:'ل'; re:'ﻠ'; join:arMedial),
  (de:'لآ'; re:'ﻵ'; join:arIsolated),
  (de:'لآ'; re:'ﻶ'; join:arFinal),
  (de:'لأ'; re:'ﻷ'; join:arIsolated),
  (de:'لأ'; re:'ﻸ'; join:arFinal),
  (de:'لإ'; re:'ﻹ'; join:arIsolated),
  (de:'لإ'; re:'ﻺ'; join:arFinal),
  (de:'لا'; re:'ﻻ'; join:arIsolated),
  (de:'لا'; re:'ﻼ'; join:arFinal),
  (de:'م'; re:'ﻡ'; join:arIsolated),
  (de:'م'; re:'ﻢ'; join:arFinal),
  (de:'م'; re:'ﻣ'; join:arInitial),
  (de:'م'; re:'ﻤ'; join:arMedial),
  (de:'ن'; re:'ﻥ'; join:arIsolated),
  (de:'ن'; re:'ﻦ'; join:arFinal),
  (de:'ن'; re:'ﻧ'; join:arInitial),
  (de:'ن'; re:'ﻨ'; join:arMedial),
  (de:'ه'; re:'ﻩ'; join:arIsolated),
  (de:'ه'; re:'ﻪ'; join:arFinal),
  (de:'ه'; re:'ﻫ'; join:arInitial),
  (de:'ه'; re:'ﻬ'; join:arMedial),
  (de:'و'; re:'ﻭ'; join:arIsolated),
  (de:'و'; re:'ﻮ'; join:arFinal),
  (de:'ؤ'; re:'ؤ'; join:arNone),
  (de:'ى'; re:'ﯨ'; join:arInitial),
  (de:'ى'; re:'ﯩ'; join:arMedial),
  (de:'ى'; re:'ﻯ'; join:arIsolated),
  (de:'ى'; re:'ﻰ'; join:arFinal),
  (de:'ي'; re:'ﻱ'; join:arIsolated),
  (de:'ي'; re:'ﻲ'; join:arFinal),
  (de:'ي'; re:'ﻳ'; join:arInitial),
  (de:'ي'; re:'ﻴ'; join:arMedial),
  (de:'ئ'; re:'ئ'; join:arNone),
  (de:'ٱ'; re:'ﭐ'; join:arIsolated),
  (de:'ٱ'; re:'ﭑ'; join:arFinal),
  (de:'ٷ'; re:'ﯝ'; join:arIsolated),
  (de:'ٹ'; re:'ﭦ'; join:arIsolated),
  (de:'ٹ'; re:'ﭧ'; join:arFinal),
  (de:'ٹ'; re:'ﭨ'; join:arInitial),
  (de:'ٹ'; re:'ﭩ'; join:arMedial),
  (de:'ٺ'; re:'ﭞ'; join:arIsolated),
  (de:'ٺ'; re:'ﭟ'; join:arFinal),
  (de:'ٺ'; re:'ﭠ'; join:arInitial),
  (de:'ٺ'; re:'ﭡ'; join:arMedial),
  (de:'ٻ'; re:'ﭒ'; join:arIsolated),
  (de:'ٻ'; re:'ﭓ'; join:arFinal),
  (de:'ٻ'; re:'ﭔ'; join:arInitial),
  (de:'ٻ'; re:'ﭕ'; join:arMedial),
  (de:'پ'; re:'ﭖ'; join:arIsolated),
  (de:'پ'; re:'ﭗ'; join:arFinal),
  (de:'پ'; re:'ﭘ'; join:arInitial),
  (de:'پ'; re:'ﭙ'; join:arMedial),
  (de:'ٿ'; re:'ﭢ'; join:arIsolated),
  (de:'ٿ'; re:'ﭣ'; join:arFinal),
  (de:'ٿ'; re:'ﭤ'; join:arInitial),
  (de:'ٿ'; re:'ﭥ'; join:arMedial),
  (de:'ڀ'; re:'ﭚ'; join:arIsolated),
  (de:'ڀ'; re:'ﭛ'; join:arFinal),
  (de:'ڀ'; re:'ﭜ'; join:arInitial),
  (de:'ڀ'; re:'ﭝ'; join:arMedial),
  (de:'ڃ'; re:'ﭶ'; join:arIsolated),
  (de:'ڃ'; re:'ﭷ'; join:arFinal),
  (de:'ڃ'; re:'ﭸ'; join:arInitial),
  (de:'ڃ'; re:'ﭹ'; join:arMedial),
  (de:'ڄ'; re:'ﭲ'; join:arIsolated),
  (de:'ڄ'; re:'ﭳ'; join:arFinal),
  (de:'ڄ'; re:'ﭴ'; join:arInitial),
  (de:'ڄ'; re:'ﭵ'; join:arMedial),
  (de:'چ'; re:'ﭺ'; join:arIsolated),
  (de:'چ'; re:'ﭻ'; join:arFinal),
  (de:'چ'; re:'ﭼ'; join:arInitial),
  (de:'چ'; re:'ﭽ'; join:arMedial),
  (de:'ڇ'; re:'ﭾ'; join:arIsolated),
  (de:'ڇ'; re:'ﭿ'; join:arFinal),
  (de:'ڇ'; re:'ﮀ'; join:arInitial),
  (de:'ڇ'; re:'ﮁ'; join:arMedial),
  (de:'ڈ'; re:'ﮈ'; join:arIsolated),
  (de:'ڈ'; re:'ﮉ'; join:arFinal),
  (de:'ڌ'; re:'ﮄ'; join:arIsolated),
  (de:'ڌ'; re:'ﮅ'; join:arFinal),
  (de:'ڍ'; re:'ﮂ'; join:arIsolated),
  (de:'ڍ'; re:'ﮃ'; join:arFinal),
  (de:'ڎ'; re:'ﮆ'; join:arIsolated),
  (de:'ڎ'; re:'ﮇ'; join:arFinal),
  (de:'ڑ'; re:'ﮌ'; join:arIsolated),
  (de:'ڑ'; re:'ﮍ'; join:arFinal),
  (de:'ژ'; re:'ﮊ'; join:arIsolated),
  (de:'ژ'; re:'ﮋ'; join:arFinal),
  (de:'ڤ'; re:'ﭪ'; join:arIsolated),
  (de:'ڤ'; re:'ﭫ'; join:arFinal),
  (de:'ڤ'; re:'ﭬ'; join:arInitial),
  (de:'ڤ'; re:'ﭭ'; join:arMedial),
  (de:'ڦ'; re:'ﭮ'; join:arIsolated),
  (de:'ڦ'; re:'ﭯ'; join:arFinal),
  (de:'ڦ'; re:'ﭰ'; join:arInitial),
  (de:'ڦ'; re:'ﭱ'; join:arMedial),
  (de:'ک'; re:'ﮎ'; join:arIsolated),
  (de:'ک'; re:'ﮏ'; join:arFinal),
  (de:'ک'; re:'ﮐ'; join:arInitial),
  (de:'ک'; re:'ﮑ'; join:arMedial),
  (de:'ڭ'; re:'ﯓ'; join:arIsolated),
  (de:'ڭ'; re:'ﯔ'; join:arFinal),
  (de:'ڭ'; re:'ﯕ'; join:arInitial),
  (de:'ڭ'; re:'ﯖ'; join:arMedial),
  (de:'گ'; re:'ﮒ'; join:arIsolated),
  (de:'گ'; re:'ﮓ'; join:arFinal),
  (de:'گ'; re:'ﮔ'; join:arInitial),
  (de:'گ'; re:'ﮕ'; join:arMedial),
  (de:'ڱ'; re:'ﮚ'; join:arIsolated),
  (de:'ڱ'; re:'ﮛ'; join:arFinal),
  (de:'ڱ'; re:'ﮜ'; join:arInitial),
  (de:'ڱ'; re:'ﮝ'; join:arMedial),
  (de:'ڳ'; re:'ﮖ'; join:arIsolated),
  (de:'ڳ'; re:'ﮗ'; join:arFinal),
  (de:'ڳ'; re:'ﮘ'; join:arInitial),
  (de:'ڳ'; re:'ﮙ'; join:arMedial),
  (de:'ں'; re:'ﮞ'; join:arIsolated),
  (de:'ں'; re:'ﮟ'; join:arFinal),
  (de:'ڻ'; re:'ﮠ'; join:arIsolated),
  (de:'ڻ'; re:'ﮡ'; join:arFinal),
  (de:'ڻ'; re:'ﮢ'; join:arInitial),
  (de:'ڻ'; re:'ﮣ'; join:arMedial),
  (de:'ھ'; re:'ﮪ'; join:arIsolated),
  (de:'ھ'; re:'ﮫ'; join:arFinal),
  (de:'ھ'; re:'ﮬ'; join:arInitial),
  (de:'ھ'; re:'ﮭ'; join:arMedial),
  (de:'ۀ'; re:'ﮤ'; join:arIsolated),
  (de:'ۀ'; re:'ﮥ'; join:arFinal),
  (de:'ہ'; re:'ﮦ'; join:arIsolated),
  (de:'ہ'; re:'ﮧ'; join:arFinal),
  (de:'ہ'; re:'ﮨ'; join:arInitial),
  (de:'ہ'; re:'ﮩ'; join:arMedial),
  (de:'ۂ'; re:'ۂ'; join:arNone),
  (de:'ۅ'; re:'ﯠ'; join:arIsolated),
  (de:'ۅ'; re:'ﯡ'; join:arFinal),
  (de:'ۆ'; re:'ﯙ'; join:arIsolated),
  (de:'ۆ'; re:'ﯚ'; join:arFinal),
  (de:'ۇ'; re:'ﯗ'; join:arIsolated),
  (de:'ۇ'; re:'ﯘ'; join:arFinal),
  (de:'ۈ'; re:'ﯛ'; join:arIsolated),
  (de:'ۈ'; re:'ﯜ'; join:arFinal),
  (de:'ۉ'; re:'ﯢ'; join:arIsolated),
  (de:'ۉ'; re:'ﯣ'; join:arFinal),
  (de:'ۋ'; re:'ﯞ'; join:arIsolated),
  (de:'ۋ'; re:'ﯟ'; join:arFinal),
  (de:'ی'; re:'ﯼ'; join:arIsolated),
  (de:'ی'; re:'ﯽ'; join:arFinal),
  (de:'ی'; re:'ﯾ'; join:arInitial),
  (de:'ی'; re:'ﯿ'; join:arMedial),
  (de:'ې'; re:'ﯤ'; join:arIsolated),
  (de:'ې'; re:'ﯥ'; join:arFinal),
  (de:'ې'; re:'ﯦ'; join:arInitial),
  (de:'ې'; re:'ﯧ'; join:arMedial),
  (de:'ے'; re:'ﮮ'; join:arIsolated),
  (de:'ے'; re:'ﮯ'; join:arFinal),
  (de:'ۓ'; re:'ۓ'; join:arNone),
  (de:'ۓ'; re:'ﮰ'; join:arIsolated),
  (de:'ۓ'; re:'ﮱ'; join:arFinal),
  (de:'ۀ'; re:'ۀ'; join:arNone),
  (de:'क़'; re:'क़'; join:arNone),
  (de:'ख़'; re:'ख़'; join:arNone),
  (de:'ग़'; re:'ग़'; join:arNone),
  (de:'ज़'; re:'ज़'; join:arNone),
  (de:'ड़'; re:'ड़'; join:arNone),
  (de:'ढ़'; re:'ढ़'; join:arNone),
  (de:'ऩ'; re:'ऩ'; join:arNone),
  (de:'फ़'; re:'फ़'; join:arNone),
  (de:'य़'; re:'य़'; join:arNone),
  (de:'ऱ'; re:'ऱ'; join:arNone),
  (de:'ऴ'; re:'ऴ'; join:arNone),
  (de:'ড়'; re:'ড়'; join:arNone),
  (de:'ঢ়'; re:'ঢ়'; join:arNone),
  (de:'য়'; re:'য়'; join:arNone),
  (de:'ਖ਼'; re:'ਖ਼'; join:arNone),
  (de:'ਗ਼'; re:'ਗ਼'; join:arNone),
  (de:'ਜ਼'; re:'ਜ਼'; join:arNone),
  (de:'ਫ਼'; re:'ਫ਼'; join:arNone),
  (de:'ਲ਼'; re:'ਲ਼'; join:arNone),
  (de:'ਸ਼'; re:'ਸ਼'; join:arNone),
  (de:'ଡ଼'; re:'ଡ଼'; join:arNone),
  (de:'ଢ଼'; re:'ଢ଼'; join:arNone),
  (de:'ୈ'; re:'ୈ'; join:arNone),
  (de:'ේ'; re:'ේ'; join:arNone),
  (de:'ෝ'; re:'ෝ'; join:arNone),
  (de:'ཀྵ'; re:'ཀྵ'; join:arNone),
  (de:'གྷ'; re:'གྷ'; join:arNone),
  (de:'ཌྷ'; re:'ཌྷ'; join:arNone),
  (de:'དྷ'; re:'དྷ'; join:arNone),
  (de:'བྷ'; re:'བྷ'; join:arNone),
  (de:'ཛྷ'; re:'ཛྷ'; join:arNone),
  (de:'ဦ'; re:'ဦ'; join:arNone),
  (de:'ᬻ'; re:'ᬻ'; join:arNone),
  (de:'ᬽ'; re:'ᬽ'; join:arNone),
  (de:'ᭃ'; re:'ᭃ'; join:arNone),
  (de:'Ḹ'; re:'Ḹ'; join:arNone),
  (de:'ḹ'; re:'ḹ'; join:arNone),
  (de:'Ṝ'; re:'Ṝ'; join:arNone),
  (de:'ṝ'; re:'ṝ'; join:arNone),
  (de:'Ṩ'; re:'Ṩ'; join:arNone),
  (de:'ṩ'; re:'ṩ'; join:arNone),
  (de:'Ậ'; re:'Ậ'; join:arNone),
  (de:'Ặ'; re:'Ặ'; join:arNone),
  (de:'ậ'; re:'ậ'; join:arNone),
  (de:'ặ'; re:'ặ'; join:arNone),
  (de:'Ệ'; re:'Ệ'; join:arNone),
  (de:'ệ'; re:'ệ'; join:arNone),
  (de:'Ộ'; re:'Ộ'; join:arNone),
  (de:'ộ'; re:'ộ'; join:arNone),
  (de:'ἂ'; re:'ἂ'; join:arNone),
  (de:'ἄ'; re:'ἄ'; join:arNone),
  (de:'ἆ'; re:'ἆ'; join:arNone),
  (de:'ᾀ'; re:'ᾀ'; join:arNone),
  (de:'ἃ'; re:'ἃ'; join:arNone),
  (de:'ἅ'; re:'ἅ'; join:arNone),
  (de:'ἇ'; re:'ἇ'; join:arNone),
  (de:'ᾁ'; re:'ᾁ'; join:arNone),
  (de:'ᾂ'; re:'ᾂ'; join:arNone),
  (de:'ᾃ'; re:'ᾃ'; join:arNone),
  (de:'ᾄ'; re:'ᾄ'; join:arNone),
  (de:'ᾅ'; re:'ᾅ'; join:arNone),
  (de:'ᾆ'; re:'ᾆ'; join:arNone),
  (de:'ᾇ'; re:'ᾇ'; join:arNone),
  (de:'Ἂ'; re:'Ἂ'; join:arNone),
  (de:'Ἄ'; re:'Ἄ'; join:arNone),
  (de:'Ἆ'; re:'Ἆ'; join:arNone),
  (de:'ᾈ'; re:'ᾈ'; join:arNone),
  (de:'Ἃ'; re:'Ἃ'; join:arNone),
  (de:'Ἅ'; re:'Ἅ'; join:arNone),
  (de:'Ἇ'; re:'Ἇ'; join:arNone),
  (de:'ᾉ'; re:'ᾉ'; join:arNone),
  (de:'ᾊ'; re:'ᾊ'; join:arNone),
  (de:'ᾋ'; re:'ᾋ'; join:arNone),
  (de:'ᾌ'; re:'ᾌ'; join:arNone),
  (de:'ᾍ'; re:'ᾍ'; join:arNone),
  (de:'ᾎ'; re:'ᾎ'; join:arNone),
  (de:'ᾏ'; re:'ᾏ'; join:arNone),
  (de:'ἒ'; re:'ἒ'; join:arNone),
  (de:'ἔ'; re:'ἔ'; join:arNone),
  (de:'ἓ'; re:'ἓ'; join:arNone),
  (de:'ἕ'; re:'ἕ'; join:arNone),
  (de:'Ἒ'; re:'Ἒ'; join:arNone),
  (de:'Ἔ'; re:'Ἔ'; join:arNone),
  (de:'Ἓ'; re:'Ἓ'; join:arNone),
  (de:'Ἕ'; re:'Ἕ'; join:arNone),
  (de:'ἢ'; re:'ἢ'; join:arNone),
  (de:'ἤ'; re:'ἤ'; join:arNone),
  (de:'ἦ'; re:'ἦ'; join:arNone),
  (de:'ᾐ'; re:'ᾐ'; join:arNone),
  (de:'ἣ'; re:'ἣ'; join:arNone),
  (de:'ἥ'; re:'ἥ'; join:arNone),
  (de:'ἧ'; re:'ἧ'; join:arNone),
  (de:'ᾑ'; re:'ᾑ'; join:arNone),
  (de:'ᾒ'; re:'ᾒ'; join:arNone),
  (de:'ᾓ'; re:'ᾓ'; join:arNone),
  (de:'ᾔ'; re:'ᾔ'; join:arNone),
  (de:'ᾕ'; re:'ᾕ'; join:arNone),
  (de:'ᾖ'; re:'ᾖ'; join:arNone),
  (de:'ᾗ'; re:'ᾗ'; join:arNone),
  (de:'Ἢ'; re:'Ἢ'; join:arNone),
  (de:'Ἤ'; re:'Ἤ'; join:arNone),
  (de:'Ἦ'; re:'Ἦ'; join:arNone),
  (de:'ᾘ'; re:'ᾘ'; join:arNone),
  (de:'Ἣ'; re:'Ἣ'; join:arNone),
  (de:'Ἥ'; re:'Ἥ'; join:arNone),
  (de:'Ἧ'; re:'Ἧ'; join:arNone),
  (de:'ᾙ'; re:'ᾙ'; join:arNone),
  (de:'ᾚ'; re:'ᾚ'; join:arNone),
  (de:'ᾛ'; re:'ᾛ'; join:arNone),
  (de:'ᾜ'; re:'ᾜ'; join:arNone),
  (de:'ᾝ'; re:'ᾝ'; join:arNone),
  (de:'ᾞ'; re:'ᾞ'; join:arNone),
  (de:'ᾟ'; re:'ᾟ'; join:arNone),
  (de:'ἲ'; re:'ἲ'; join:arNone),
  (de:'ἴ'; re:'ἴ'; join:arNone),
  (de:'ἶ'; re:'ἶ'; join:arNone),
  (de:'ἳ'; re:'ἳ'; join:arNone),
  (de:'ἵ'; re:'ἵ'; join:arNone),
  (de:'ἷ'; re:'ἷ'; join:arNone),
  (de:'Ἲ'; re:'Ἲ'; join:arNone),
  (de:'Ἴ'; re:'Ἴ'; join:arNone),
  (de:'Ἶ'; re:'Ἶ'; join:arNone),
  (de:'Ἳ'; re:'Ἳ'; join:arNone),
  (de:'Ἵ'; re:'Ἵ'; join:arNone),
  (de:'Ἷ'; re:'Ἷ'; join:arNone),
  (de:'ὂ'; re:'ὂ'; join:arNone),
  (de:'ὄ'; re:'ὄ'; join:arNone),
  (de:'ὃ'; re:'ὃ'; join:arNone),
  (de:'ὅ'; re:'ὅ'; join:arNone),
  (de:'Ὂ'; re:'Ὂ'; join:arNone),
  (de:'Ὄ'; re:'Ὄ'; join:arNone),
  (de:'Ὃ'; re:'Ὃ'; join:arNone),
  (de:'Ὅ'; re:'Ὅ'; join:arNone),
  (de:'ὒ'; re:'ὒ'; join:arNone),
  (de:'ὔ'; re:'ὔ'; join:arNone),
  (de:'ὖ'; re:'ὖ'; join:arNone),
  (de:'ὓ'; re:'ὓ'; join:arNone),
  (de:'ὕ'; re:'ὕ'; join:arNone),
  (de:'ὗ'; re:'ὗ'; join:arNone),
  (de:'Ὓ'; re:'Ὓ'; join:arNone),
  (de:'Ὕ'; re:'Ὕ'; join:arNone),
  (de:'Ὗ'; re:'Ὗ'; join:arNone),
  (de:'ὢ'; re:'ὢ'; join:arNone),
  (de:'ὤ'; re:'ὤ'; join:arNone),
  (de:'ὦ'; re:'ὦ'; join:arNone),
  (de:'ᾠ'; re:'ᾠ'; join:arNone),
  (de:'ὣ'; re:'ὣ'; join:arNone),
  (de:'ὥ'; re:'ὥ'; join:arNone),
  (de:'ὧ'; re:'ὧ'; join:arNone),
  (de:'ᾡ'; re:'ᾡ'; join:arNone),
  (de:'ᾢ'; re:'ᾢ'; join:arNone),
  (de:'ᾣ'; re:'ᾣ'; join:arNone),
  (de:'ᾤ'; re:'ᾤ'; join:arNone),
  (de:'ᾥ'; re:'ᾥ'; join:arNone),
  (de:'ᾦ'; re:'ᾦ'; join:arNone),
  (de:'ᾧ'; re:'ᾧ'; join:arNone),
  (de:'Ὢ'; re:'Ὢ'; join:arNone),
  (de:'Ὤ'; re:'Ὤ'; join:arNone),
  (de:'Ὦ'; re:'Ὦ'; join:arNone),
  (de:'ᾨ'; re:'ᾨ'; join:arNone),
  (de:'Ὣ'; re:'Ὣ'; join:arNone),
  (de:'Ὥ'; re:'Ὥ'; join:arNone),
  (de:'Ὧ'; re:'Ὧ'; join:arNone),
  (de:'ᾩ'; re:'ᾩ'; join:arNone),
  (de:'ᾪ'; re:'ᾪ'; join:arNone),
  (de:'ᾫ'; re:'ᾫ'; join:arNone),
  (de:'ᾬ'; re:'ᾬ'; join:arNone),
  (de:'ᾭ'; re:'ᾭ'; join:arNone),
  (de:'ᾮ'; re:'ᾮ'; join:arNone),
  (de:'ᾯ'; re:'ᾯ'; join:arNone),
  (de:'ᾲ'; re:'ᾲ'; join:arNone),
  (de:'ῂ'; re:'ῂ'; join:arNone),
  (de:'ῲ'; re:'ῲ'; join:arNone),
  (de:'ᾷ'; re:'ᾷ'; join:arNone),
  (de:'῍'; re:'῍'; join:arNone),
  (de:'῎'; re:'῎'; join:arNone),
  (de:'῏'; re:'῏'; join:arNone),
  (de:'ῇ'; re:'ῇ'; join:arNone),
  (de:'ῷ'; re:'ῷ'; join:arNone),
  (de:'῝'; re:'῝'; join:arNone),
  (de:'῞'; re:'῞'; join:arNone),
  (de:'῟'; re:'῟'; join:arNone),
  (de:'↚'; re:'↚'; join:arNone),
  (de:'↛'; re:'↛'; join:arNone),
  (de:'↮'; re:'↮'; join:arNone),
  (de:'⇍'; re:'⇍'; join:arNone),
  (de:'⇏'; re:'⇏'; join:arNone),
  (de:'⇎'; re:'⇎'; join:arNone),
  (de:'∄'; re:'∄'; join:arNone),
  (de:'∉'; re:'∉'; join:arNone),
  (de:'∌'; re:'∌'; join:arNone),
  (de:'∤'; re:'∤'; join:arNone),
  (de:'∦'; re:'∦'; join:arNone),
  (de:'≁'; re:'≁'; join:arNone),
  (de:'≄'; re:'≄'; join:arNone),
  (de:'≇'; re:'≇'; join:arNone),
  (de:'≉'; re:'≉'; join:arNone),
  (de:'≭'; re:'≭'; join:arNone),
  (de:'≢'; re:'≢'; join:arNone),
  (de:'≰'; re:'≰'; join:arNone),
  (de:'≱'; re:'≱'; join:arNone),
  (de:'≴'; re:'≴'; join:arNone),
  (de:'≵'; re:'≵'; join:arNone),
  (de:'≸'; re:'≸'; join:arNone),
  (de:'≹'; re:'≹'; join:arNone),
  (de:'⊀'; re:'⊀'; join:arNone),
  (de:'⊁'; re:'⊁'; join:arNone),
  (de:'⋠'; re:'⋠'; join:arNone),
  (de:'⋡'; re:'⋡'; join:arNone),
  (de:'⊄'; re:'⊄'; join:arNone),
  (de:'⊅'; re:'⊅'; join:arNone),
  (de:'⊈'; re:'⊈'; join:arNone),
  (de:'⊉'; re:'⊉'; join:arNone),
  (de:'⋢'; re:'⋢'; join:arNone),
  (de:'⋣'; re:'⋣'; join:arNone),
  (de:'⊬'; re:'⊬'; join:arNone),
  (de:'⊭'; re:'⊭'; join:arNone),
  (de:'⊮'; re:'⊮'; join:arNone),
  (de:'⊯'; re:'⊯'; join:arNone),
  (de:'⋪'; re:'⋪'; join:arNone),
  (de:'⋫'; re:'⋫'; join:arNone),
  (de:'⋬'; re:'⋬'; join:arNone),
  (de:'⋭'; re:'⋭'; join:arNone),
  (de:'⫝̸'; re:'⫝̸'; join:arNone),
  (de:'ゔ'; re:'ゔ'; join:arNone),
  (de:'が'; re:'が'; join:arNone),
  (de:'ぎ'; re:'ぎ'; join:arNone),
  (de:'ぐ'; re:'ぐ'; join:arNone),
  (de:'げ'; re:'げ'; join:arNone),
  (de:'ご'; re:'ご'; join:arNone),
  (de:'ざ'; re:'ざ'; join:arNone),
  (de:'じ'; re:'じ'; join:arNone),
  (de:'ず'; re:'ず'; join:arNone),
  (de:'ぜ'; re:'ぜ'; join:arNone),
  (de:'ぞ'; re:'ぞ'; join:arNone),
  (de:'だ'; re:'だ'; join:arNone),
  (de:'ぢ'; re:'ぢ'; join:arNone),
  (de:'づ'; re:'づ'; join:arNone),
  (de:'で'; re:'で'; join:arNone),
  (de:'ど'; re:'ど'; join:arNone),
  (de:'ば'; re:'ば'; join:arNone),
  (de:'ぱ'; re:'ぱ'; join:arNone),
  (de:'び'; re:'び'; join:arNone),
  (de:'ぴ'; re:'ぴ'; join:arNone),
  (de:'ぶ'; re:'ぶ'; join:arNone),
  (de:'ぷ'; re:'ぷ'; join:arNone),
  (de:'べ'; re:'べ'; join:arNone),
  (de:'ぺ'; re:'ぺ'; join:arNone),
  (de:'ぼ'; re:'ぼ'; join:arNone),
  (de:'ぽ'; re:'ぽ'; join:arNone),
  (de:'ゞ'; re:'ゞ'; join:arNone),
  (de:'ヴ'; re:'ヴ'; join:arNone),
  (de:'ガ'; re:'ガ'; join:arNone),
  (de:'ギ'; re:'ギ'; join:arNone),
  (de:'グ'; re:'グ'; join:arNone),
  (de:'ゲ'; re:'ゲ'; join:arNone),
  (de:'ゴ'; re:'ゴ'; join:arNone),
  (de:'ザ'; re:'ザ'; join:arNone),
  (de:'ジ'; re:'ジ'; join:arNone),
  (de:'ズ'; re:'ズ'; join:arNone),
  (de:'ゼ'; re:'ゼ'; join:arNone),
  (de:'ゾ'; re:'ゾ'; join:arNone),
  (de:'ダ'; re:'ダ'; join:arNone),
  (de:'ヂ'; re:'ヂ'; join:arNone),
  (de:'ヅ'; re:'ヅ'; join:arNone),
  (de:'デ'; re:'デ'; join:arNone),
  (de:'ド'; re:'ド'; join:arNone),
  (de:'バ'; re:'バ'; join:arNone),
  (de:'パ'; re:'パ'; join:arNone),
  (de:'ビ'; re:'ビ'; join:arNone),
  (de:'ピ'; re:'ピ'; join:arNone),
  (de:'ブ'; re:'ブ'; join:arNone),
  (de:'プ'; re:'プ'; join:arNone),
  (de:'ベ'; re:'ベ'; join:arNone),
  (de:'ペ'; re:'ペ'; join:arNone),
  (de:'ボ'; re:'ボ'; join:arNone),
  (de:'ポ'; re:'ポ'; join:arNone),
  (de:'ヷ'; re:'ヷ'; join:arNone),
  (de:'ヸ'; re:'ヸ'; join:arNone),
  (de:'ヹ'; re:'ヹ'; join:arNone),
  (de:'ヺ'; re:'ヺ'; join:arNone),
  (de:'ヾ'; re:'ヾ'; join:arNone),
  (de:'שּׁ'; re:'שּׁ'; join:arNone),
  (de:'שּׂ'; re:'שּׂ'; join:arNone),
  (de:'𑂚'; re:'𑂚'; join:arNone),
  (de:'𑂜'; re:'𑂜'; join:arNone),
  (de:'𑂫'; re:'𑂫'; join:arNone),
  (de:'𑒻'; re:'𑒻'; join:arNone)
  );

